/*******************************************************************************
* DISCLAIMER
* This software is supplied by Renesas Electronics Corporation and is only 
* intended for use with Renesas products. No other uses are authorized. This 
* software is owned by Renesas Electronics Corporation and is protected under
* all applicable laws, including copyright laws.
* THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
* THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
* LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
* AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
* TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS 
* ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE 
* FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
* ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
* BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
* Renesas reserves the right, without notice, to make changes to this software
* and to discontinue the availability of this software. By using this software,
* you agree to the additional terms and conditions found by accessing the 
* following link:
* http://www.renesas.com/disclaimer 
*
* Copyright (C) 2012 Renesas Electronics Corporation. All rights reserved.    
*******************************************************************************/
/******************************************************************************
* File Name    : r_flash_rx210.h
* Description  : This file has specific information about the ROM and DF on 
*                RX210 Group MCUs.
*******************************************************************************
* History : DD.MM.YYYY Version Description
*         : 15.10.2012 1.00    First Release
*         : 01.12.2015 1.10    Fixed incorrect FLASH_CF_BLOCK_INVALID value
*         : 09.08.2016 2.00    Modified for BSPless flash
*         : 17.11.2016 2.10    Added FLASH_xxx size and num block defines
*                              Fixed FLASH_CF_BLOCK_INVALID for 1M parts.
******************************************************************************/

#ifndef _FLASH_API_RX210_H
#define _FLASH_API_RX210_H

/******************************************************************************
Includes   <System Includes> , "Project Includes"
******************************************************************************/
/* Defines standard typedefs used in this file */
#include "r_mcu_config.h"

#include <stdint.h>

/******************************************************************************
Macro definitions
******************************************************************************/
/* Defines the number of flash areas */
#define NUM_ROM_AREAS           1
/* Defines the start program/erase address for the different flash areas */
#define ROM_AREA_0              (0x00F80000)
   
/* Defines whether this MCU requires the FCU RAM be enabled and initialized.
   If uncommented, then MCU is required to init FCU RAM.
   If commented out, then MCU is not required. */
#define FCU_RAM_INIT_REQUIRED   (1)

/*  Bottom of DF Area */
#define DF_ADDRESS              0x00100000
/* Used for getting DF block */
#define DF_MASK                 0xFFFFF800
/* Used for getting erase boundary in DF block when doing blank checking */
#define DF_ERASE_BLOCK_SIZE     0x00000080
/* Defines how many DF blocks are on this part */
#define DF_NUM_BLOCKS           64      // 64 blocks 0f 128 = 8K
/* THis is the last address in the last block, not the starting address */
#define FLASH_DF_LAST_BLOCK        ((FLASH_DF_BLOCK_INVALID) - (DF_ERASE_BLOCK_SIZE))

/* Defines how many ROM blocks are on this part */
#if MCU_ROM_SIZE_BYTES == 65536
    #define ROM_NUM_BLOCKS  32        //64KB part
#elif MCU_ROM_SIZE_BYTES == 98304
    #define ROM_NUM_BLOCKS  48        //96KB part
#elif MCU_ROM_SIZE_BYTES == 131072
    #define ROM_NUM_BLOCKS  64    	  //128KB part
#elif MCU_ROM_SIZE_BYTES == 262144
    #define ROM_NUM_BLOCKS  128  	  //256KB part
#elif MCU_ROM_SIZE_BYTES == 393216
    #define ROM_NUM_BLOCKS  192  	  //384KB part
#elif MCU_ROM_SIZE_BYTES == 524288
    #define ROM_NUM_BLOCKS  256  	  //512KB part
#elif MCU_ROM_SIZE_BYTES == 786432
    #define ROM_NUM_BLOCKS  384       //512KB part
#elif MCU_ROM_SIZE_BYTES == 1048576
    #define ROM_NUM_BLOCKS  512       //1MB part
#else
    #error "ERROR - Flash API - This ROM size is not defined. Update r_flash_rx210.h with this MCUs information." 
#endif

/* Programming size for ROM in bytes. This is being set as 2 to comply with older versions where there was only 1
   size option.  */
#define ROM_PROGRAM_SIZE          2
/* Programming size for ROM in bytes. RX210 support 128, 8 and 2 byte options. */
#define ROM_PROGRAM_SIZE_LARGE    128
#define ROM_PROGRAM_SIZE_MEDIUM   8
#define ROM_PROGRAM_SIZE_SMALL    2
/* Programming size for data flash in bytes */
/* NOTE: RX210 can program 2 or 8 bytes at a time. */
#define DF_PROGRAM_SIZE_LARGE     8
#define DF_PROGRAM_SIZE_SMALL     2

#define FLASH_CF_BLOCK_SIZE             (2048)
#define FLASH_NUM_BLOCKS_CF             ROM_NUM_BLOCKS
#define FLASH_CF_MIN_PGM_SIZE           ROM_PROGRAM_SIZE
#define FLASH_DF_BLOCK_SIZE             DF_ERASE_BLOCK_SIZE
#define FLASH_NUM_BLOCKS_DF             DF_NUM_BLOCKS
#define FLASH_DF_MIN_PGM_SIZE           DF_PROGRAM_SIZE_SMALL
#define FLASH_DF_HIGHEST_VALID_BLOCK    FLASH_DF_LAST_BLOCK


/* User ROM Block Area        Size: Start Addr -   End Addr */
typedef enum _flash_block_address
{
    FLASH_CF_BLOCK_0     = 0x00FFF800,    /*   2KB: 0xFFFFF800 - 0xFFFFFFFF */
    FLASH_CF_BLOCK_1     = 0x00FFF000,    /*   2KB: 0xFFFFF000 - 0xFFFFF7FF */
    FLASH_CF_BLOCK_2     = 0x00FFE800,    /*   2KB: 0xFFFFE800 - 0xFFFFEFFF */
    FLASH_CF_BLOCK_3     = 0x00FFE000,    /*   2KB: 0xFFFFE000 - 0xFFFFE7FF */
    FLASH_CF_BLOCK_4     = 0xFFFFD800,    /*   2KB: 0xFFFFD800 - 0xFFFFDFFF */
    FLASH_CF_BLOCK_5     = 0x00FFD000,    /*   2KB: 0xFFFFD000 - 0xFFFFD7FF */
    FLASH_CF_BLOCK_6     = 0x00FFC800,    /*   2KB: 0xFFFFC800 - 0xFFFFCFFF */
    FLASH_CF_BLOCK_7     = 0x00FFC000,    /*   2KB: 0xFFFFC000 - 0xFFFFC7FF */
    FLASH_CF_BLOCK_8     = 0x00FFB800,    /*   2KB: 0xFFFFB800 - 0xFFFFBFFF */
    FLASH_CF_BLOCK_9     = 0x00FFB000,    /*   2KB: 0xFFFFB000 - 0xFFFFB7FF */
    FLASH_CF_BLOCK_10    = 0x00FFA800,    /*   2KB: 0xFFFFA800 - 0xFFFFAFFF */
    FLASH_CF_BLOCK_11    = 0x00FFA000,    /*   2KB: 0xFFFFA000 - 0xFFFFA7FF */
    FLASH_CF_BLOCK_12    = 0x00FF9800,    /*   2KB: 0xFFFF9800 - 0xFFFF9FFF */
    FLASH_CF_BLOCK_13    = 0x00FF9000,    /*   2KB: 0xFFFF9000 - 0xFFFF97FF */
    FLASH_CF_BLOCK_14    = 0x00FF8800,    /*   2KB: 0xFFFF8800 - 0xFFFF8FFF */
    FLASH_CF_BLOCK_15    = 0x00FF8000,    /*   2KB: 0xFFFF8000 - 0xFFFF87FF */
    FLASH_CF_BLOCK_16    = 0x00FF7800,    /*   2KB: 0xFFFF7800 - 0xFFFF7FFF */
    FLASH_CF_BLOCK_17    = 0x00FF7000,    /*   2KB: 0xFFFF7000 - 0xFFFF77FF */
    FLASH_CF_BLOCK_18    = 0x00FF6800,    /*   2KB: 0xFFFF6800 - 0xFFFF6FFF */
    FLASH_CF_BLOCK_19    = 0x00FF6000,    /*   2KB: 0xFFFF6000 - 0xFFFF67FF */
    FLASH_CF_BLOCK_20    = 0x00FF5800,    /*   2KB: 0xFFFF5800 - 0xFFFF5FFF */
    FLASH_CF_BLOCK_21    = 0x00FF5000,    /*   2KB: 0xFFFF5000 - 0xFFFF57FF */
    FLASH_CF_BLOCK_22    = 0x00FF4800,    /*   2KB: 0xFFFF4800 - 0xFFFF4FFF */
    FLASH_CF_BLOCK_23    = 0x00FF4000,    /*   2KB: 0xFFFF4000 - 0xFFFF47FF */
    FLASH_CF_BLOCK_24    = 0x00FF3800,    /*   2KB: 0xFFFF3800 - 0xFFFF3FFF */
    FLASH_CF_BLOCK_25    = 0x00FF3000,    /*   2KB: 0xFFFF3000 - 0xFFFF37FF */
    FLASH_CF_BLOCK_26    = 0x00FF2800,    /*   2KB: 0xFFFF2800 - 0xFFFF2FFF */
    FLASH_CF_BLOCK_27    = 0x00FF2000,    /*   2KB: 0xFFFF2000 - 0xFFFF27FF */
    FLASH_CF_BLOCK_28    = 0x00FF1800,    /*   2KB: 0xFFFF1800 - 0xFFFF1FFF */
    FLASH_CF_BLOCK_29    = 0x00FF1000,    /*   2KB: 0xFFFF1000 - 0xFFFF17FF */
    FLASH_CF_BLOCK_30    = 0x00FF0800,    /*   2KB: 0xFFFF0800 - 0xFFFF0FFF */
    FLASH_CF_BLOCK_31    = 0x00FF0000,    /*   2KB: 0xFFFF0000 - 0xFFFF07FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x3   //   64KB ROM
    FLASH_CF_BLOCK_INVALID = 0x00FEFFFF,
#else
    FLASH_CF_BLOCK_32    = 0x00FEF800,    /*   2KB: 0xFFFEF800 - 0xFFFEFFFF */
    FLASH_CF_BLOCK_33    = 0x00FEF000,    /*   2KB: 0xFFFEF000 - 0xFFFEF7FF */
    FLASH_CF_BLOCK_34    = 0x00FEE800,    /*   2KB: 0xFFFEE800 - 0xFFFEEFFF */
    FLASH_CF_BLOCK_35    = 0x00FEE000,    /*   2KB: 0xFFFEE000 - 0xFFFEE7FF */
    FLASH_CF_BLOCK_36    = 0x00FED800,    /*   2KB: 0xFFFED800 - 0xFFFEDFFF */
    FLASH_CF_BLOCK_37    = 0x00FED000,    /*   2KB: 0xFFFED000 - 0xFFFED7FF */
    FLASH_CF_BLOCK_38    = 0x00FEC800,    /*   2KB: 0xFFFEC800 - 0xFFFECFFF */
    FLASH_CF_BLOCK_39    = 0x00FEC000,    /*   2KB: 0xFFFEC000 - 0xFFFEC7FF */
    FLASH_CF_BLOCK_40    = 0x00FEB800,    /*   2KB: 0xFFFEB800 - 0xFFFEBFFF */
    FLASH_CF_BLOCK_41    = 0x00FEB000,    /*   2KB: 0xFFFEB000 - 0xFFFEB7FF */
    FLASH_CF_BLOCK_42    = 0x00FEA800,    /*   2KB: 0xFFFEA800 - 0xFFFEAFFF */
    FLASH_CF_BLOCK_43    = 0x00FEA000,    /*   2KB: 0xFFFEA000 - 0xFFFEA7FF */
    FLASH_CF_BLOCK_44    = 0x00FE9800,    /*   2KB: 0xFFFE9800 - 0xFFFE9FFF */
    FLASH_CF_BLOCK_45    = 0x00FE9000,    /*   2KB: 0xFFFE9000 - 0xFFFE97FF */
    FLASH_CF_BLOCK_46    = 0x00FE8800,    /*   2KB: 0xFFFE8800 - 0xFFFE8FFF */
    FLASH_CF_BLOCK_47    = 0x00FE8000,    /*   2KB: 0xFFFE8000 - 0xFFFE87FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x4   //   96KB ROM
    FLASH_CF_BLOCK_INVALID = 0x00FE7FFF,
#else
    FLASH_CF_BLOCK_48    = 0x00FE7800,    /*   2KB: 0xFFFE7800 - 0xFFFE7FFF */
    FLASH_CF_BLOCK_49    = 0x00FE7000,    /*   2KB: 0xFFFE7000 - 0xFFFE77FF */
    FLASH_CF_BLOCK_50    = 0x00FE6800,    /*   2KB: 0xFFFE6800 - 0xFFFE6FFF */
    FLASH_CF_BLOCK_51    = 0x00FE6000,    /*   2KB: 0xFFFE6000 - 0xFFFE67FF */
    FLASH_CF_BLOCK_52    = 0x00FE5800,    /*   2KB: 0xFFFE5800 - 0xFFFE5FFF */
    FLASH_CF_BLOCK_53    = 0x00FE5000,    /*   2KB: 0xFFFE5000 - 0xFFFE57FF */
    FLASH_CF_BLOCK_54    = 0x00FE4800,    /*   2KB: 0xFFFE4800 - 0xFFFE4FFF */
    FLASH_CF_BLOCK_55    = 0x00FE4000,    /*   2KB: 0xFFFE4000 - 0xFFFE47FF */
    FLASH_CF_BLOCK_56    = 0x00FE3800,    /*   2KB: 0xFFFE3800 - 0xFFFE3FFF */
    FLASH_CF_BLOCK_57    = 0x00FE3000,    /*   2KB: 0xFFFE3000 - 0xFFFE37FF */
    FLASH_CF_BLOCK_58    = 0x00FE2800,    /*   2KB: 0xFFFE2800 - 0xFFFE2FFF */
    FLASH_CF_BLOCK_59    = 0x00FE2000,    /*   2KB: 0xFFFE2000 - 0xFFFE27FF */
    FLASH_CF_BLOCK_60    = 0x00FE1800,    /*   2KB: 0xFFFE1800 - 0xFFFE1FFF */
    FLASH_CF_BLOCK_61    = 0x00FE1000,    /*   2KB: 0xFFFE1000 - 0xFFFE17FF */
    FLASH_CF_BLOCK_62    = 0x00FE0800,    /*   2KB: 0xFFFE0800 - 0xFFFE0FFF */
    FLASH_CF_BLOCK_63    = 0x00FE0000,    /*   2KB: 0xFFFE0000 - 0xFFFE07FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x5   //   128KB ROM
    FLASH_CF_BLOCK_INVALID = 0x00FDFFFF,
#else
    FLASH_CF_BLOCK_64    = 0x00FDF800,    /*   2KB: 0xFFFDF800 - 0xFFFDFFFF */
    FLASH_CF_BLOCK_65    = 0x00FDF000,    /*   2KB: 0xFFFDF000 - 0xFFFDF7FF */
    FLASH_CF_BLOCK_66    = 0x00FDE800,    /*   2KB: 0xFFFDE800 - 0xFFFDEFFF */
    FLASH_CF_BLOCK_67    = 0x00FDE000,    /*   2KB: 0xFFFDE000 - 0xFFFDE7FF */
    FLASH_CF_BLOCK_68    = 0x00FDD800,    /*   2KB: 0xFFFDD800 - 0xFFFDDFFF */
    FLASH_CF_BLOCK_69    = 0x00FDD000,    /*   2KB: 0xFFFDD000 - 0xFFFDD7FF */
    FLASH_CF_BLOCK_70    = 0x00FDC800,    /*   2KB: 0xFFFDC800 - 0xFFFDCFFF */
    FLASH_CF_BLOCK_71    = 0x00FDC000,    /*   2KB: 0xFFFDC000 - 0xFFFDC7FF */
    FLASH_CF_BLOCK_72    = 0x00FDB800,    /*   2KB: 0xFFFDB800 - 0xFFFDBFFF */
    FLASH_CF_BLOCK_73    = 0x00FDB000,    /*   2KB: 0xFFFDB000 - 0xFFFDB7FF */
    FLASH_CF_BLOCK_74    = 0x00FDA800,    /*   2KB: 0xFFFDA800 - 0xFFFDAFFF */
    FLASH_CF_BLOCK_75    = 0x00FDA000,    /*   2KB: 0xFFFDA000 - 0xFFFDA7FF */
    FLASH_CF_BLOCK_76    = 0x00FD9800,    /*   2KB: 0xFFFD9800 - 0xFFFD9FFF */
    FLASH_CF_BLOCK_77    = 0x00FD9000,    /*   2KB: 0xFFFD9000 - 0xFFFD97FF */
    FLASH_CF_BLOCK_78    = 0x00FD8800,    /*   2KB: 0xFFFD8800 - 0xFFFD8FFF */
    FLASH_CF_BLOCK_79    = 0x00FD8000,    /*   2KB: 0xFFFD8000 - 0xFFFD87FF */
    FLASH_CF_BLOCK_80    = 0x00FD7800,    /*   2KB: 0xFFFD7800 - 0xFFFD7FFF */
    FLASH_CF_BLOCK_81    = 0x00FD7000,    /*   2KB: 0xFFFD7000 - 0xFFFD77FF */
    FLASH_CF_BLOCK_82    = 0x00FD6800,    /*   2KB: 0xFFFD6800 - 0xFFFD6FFF */
    FLASH_CF_BLOCK_83    = 0x00FD6000,    /*   2KB: 0xFFFD6000 - 0xFFFD67FF */
    FLASH_CF_BLOCK_84    = 0x00FD5800,    /*   2KB: 0xFFFD5800 - 0xFFFD5FFF */
    FLASH_CF_BLOCK_85    = 0x00FD5000,    /*   2KB: 0xFFFD5000 - 0xFFFD57FF */
    FLASH_CF_BLOCK_86    = 0x00FD4800,    /*   2KB: 0xFFFD4800 - 0xFFFD4FFF */
    FLASH_CF_BLOCK_87    = 0x00FD4000,    /*   2KB: 0xFFFD4000 - 0xFFFD47FF */
    FLASH_CF_BLOCK_88    = 0x00FD3800,    /*   2KB: 0xFFFD3800 - 0xFFFD3FFF */
    FLASH_CF_BLOCK_89    = 0x00FD3000,    /*   2KB: 0xFFFD3000 - 0xFFFD37FF */
    FLASH_CF_BLOCK_90    = 0x00FD2800,    /*   2KB: 0xFFFD2800 - 0xFFFD2FFF */
    FLASH_CF_BLOCK_91    = 0x00FD2000,    /*   2KB: 0xFFFD2000 - 0xFFFD27FF */
    FLASH_CF_BLOCK_92    = 0x00FD1800,    /*   2KB: 0xFFFD1800 - 0xFFFD1FFF */
    FLASH_CF_BLOCK_93    = 0x00FD1000,    /*   2KB: 0xFFFD1000 - 0xFFFD17FF */
    FLASH_CF_BLOCK_94    = 0x00FD0800,    /*   2KB: 0xFFFD0800 - 0xFFFD0FFF */
    FLASH_CF_BLOCK_95    = 0x00FD0000,    /*   2KB: 0xFFFD0000 - 0xFFFD07FF */
    FLASH_CF_BLOCK_96    = 0x00FCF800,    /*   2KB: 0xFFFCF800 - 0xFFFCFFFF */
    FLASH_CF_BLOCK_97    = 0x00FCF000,    /*   2KB: 0xFFFCF000 - 0xFFFCF7FF */
    FLASH_CF_BLOCK_98    = 0x00FCE800,    /*   2KB: 0xFFFCE800 - 0xFFFCEFFF */
    FLASH_CF_BLOCK_99    = 0x00FCE000,    /*   2KB: 0xFFFCE000 - 0xFFFCE7FF */
    FLASH_CF_BLOCK_100   = 0x00FCD800,    /*   2KB: 0xFFFCD800 - 0xFFFCDFFF */
    FLASH_CF_BLOCK_101   = 0x00FCD000,    /*   2KB: 0xFFFCD000 - 0xFFFCD7FF */
    FLASH_CF_BLOCK_102   = 0x00FCC800,    /*   2KB: 0xFFFCC800 - 0xFFFCCFFF */
    FLASH_CF_BLOCK_103   = 0x00FCC000,    /*   2KB: 0xFFFCC000 - 0xFFFCC7FF */
    FLASH_CF_BLOCK_104   = 0x00FCB800,    /*   2KB: 0xFFFCB800 - 0xFFFCBFFF */
    FLASH_CF_BLOCK_105   = 0x00FCB000,    /*   2KB: 0xFFFCB000 - 0xFFFCB7FF */
    FLASH_CF_BLOCK_106   = 0x00FCA800,    /*   2KB: 0xFFFCA800 - 0xFFFCAFFF */
    FLASH_CF_BLOCK_107   = 0x00FCA000,    /*   2KB: 0xFFFCA000 - 0xFFFCA7FF */
    FLASH_CF_BLOCK_108   = 0x00FC9800,    /*   2KB: 0xFFFC9800 - 0xFFFC9FFF */
    FLASH_CF_BLOCK_109   = 0x00FC9000,    /*   2KB: 0xFFFC9000 - 0xFFFC97FF */
    FLASH_CF_BLOCK_110   = 0x00FC8800,    /*   2KB: 0xFFFC8800 - 0xFFFC8FFF */
    FLASH_CF_BLOCK_111   = 0x00FC8000,    /*   2KB: 0xFFFC8000 - 0xFFFC87FF */
    FLASH_CF_BLOCK_112   = 0x00FC7800,    /*   2KB: 0xFFFC7800 - 0xFFFC7FFF */
    FLASH_CF_BLOCK_113   = 0x00FC7000,    /*   2KB: 0xFFFC7000 - 0xFFFC77FF */
    FLASH_CF_BLOCK_114   = 0x00FC6800,    /*   2KB: 0xFFFC6800 - 0xFFFC6FFF */
    FLASH_CF_BLOCK_115   = 0x00FC6000,    /*   2KB: 0xFFFC6000 - 0xFFFC67FF */
    FLASH_CF_BLOCK_116   = 0x00FC5800,    /*   2KB: 0xFFFC5800 - 0xFFFC5FFF */
    FLASH_CF_BLOCK_117   = 0x00FC5000,    /*   2KB: 0xFFFC5000 - 0xFFFC57FF */
    FLASH_CF_BLOCK_118   = 0x00FC4800,    /*   2KB: 0xFFFC4800 - 0xFFFC4FFF */
    FLASH_CF_BLOCK_119   = 0x00FC4000,    /*   2KB: 0xFFFC4000 - 0xFFFC47FF */
    FLASH_CF_BLOCK_120   = 0x00FC3800,    /*   2KB: 0xFFFC3800 - 0xFFFC3FFF */
    FLASH_CF_BLOCK_121   = 0x00FC3000,    /*   2KB: 0xFFFC3000 - 0xFFFC37FF */
    FLASH_CF_BLOCK_122   = 0x00FC2800,    /*   2KB: 0xFFFC2800 - 0xFFFC2FFF */
    FLASH_CF_BLOCK_123   = 0x00FC2000,    /*   2KB: 0xFFFC2000 - 0xFFFC27FF */
    FLASH_CF_BLOCK_124   = 0x00FC1800,    /*   2KB: 0xFFFC1800 - 0xFFFC1FFF */
    FLASH_CF_BLOCK_125   = 0x00FC1000,    /*   2KB: 0xFFFC1000 - 0xFFFC17FF */
    FLASH_CF_BLOCK_126   = 0x00FC0800,    /*   2KB: 0xFFFC0800 - 0xFFFC0FFF */
    FLASH_CF_BLOCK_127   = 0x00FC0000,    /*   2KB: 0xFFFC0000 - 0xFFFC07FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x6   //   256KB ROM
    FLASH_CF_BLOCK_INVALID = 0x00FBFFFF,
#else
    FLASH_CF_BLOCK_128   = 0x00FBF800,    /*   2KB: 0xFFFBF800 - 0xFFFBFFFF */
    FLASH_CF_BLOCK_129   = 0x00FBF000,    /*   2KB: 0xFFFBF000 - 0xFFFBF7FF */
    FLASH_CF_BLOCK_130   = 0x00FBE800,    /*   2KB: 0xFFFBE800 - 0xFFFBEFFF */
    FLASH_CF_BLOCK_131   = 0x00FBE000,    /*   2KB: 0xFFFBE000 - 0xFFFBE7FF */
    FLASH_CF_BLOCK_132   = 0x00FBD800,    /*   2KB: 0xFFFBD800 - 0xFFFBDFFF */
    FLASH_CF_BLOCK_133   = 0x00FBD000,    /*   2KB: 0xFFFBD000 - 0xFFFBD7FF */
    FLASH_CF_BLOCK_134   = 0x00FBC800,    /*   2KB: 0xFFFBC800 - 0xFFFBCFFF */
    FLASH_CF_BLOCK_135   = 0x00FBC000,    /*   2KB: 0xFFFBC000 - 0xFFFBC7FF */
    FLASH_CF_BLOCK_136   = 0x00FBB800,    /*   2KB: 0xFFFBB800 - 0xFFFBBFFF */
    FLASH_CF_BLOCK_137   = 0x00FBB000,    /*   2KB: 0xFFFBB000 - 0xFFFBB7FF */
    FLASH_CF_BLOCK_138   = 0x00FBA800,    /*   2KB: 0xFFFBA800 - 0xFFFBAFFF */
    FLASH_CF_BLOCK_139   = 0x00FBA000,    /*   2KB: 0xFFFBA000 - 0xFFFBA7FF */
    FLASH_CF_BLOCK_140   = 0x00FB9800,    /*   2KB: 0xFFFB9800 - 0xFFFB9FFF */
    FLASH_CF_BLOCK_141   = 0x00FB9000,    /*   2KB: 0xFFFB9000 - 0xFFFB97FF */
    FLASH_CF_BLOCK_142   = 0x00FB8800,    /*   2KB: 0xFFFB8800 - 0xFFFB8FFF */
    FLASH_CF_BLOCK_143   = 0x00FB8000,    /*   2KB: 0xFFFB8000 - 0xFFFB87FF */
    FLASH_CF_BLOCK_144   = 0x00FB7800,    /*   2KB: 0xFFFB7800 - 0xFFFB7FFF */
    FLASH_CF_BLOCK_145   = 0x00FB7000,    /*   2KB: 0xFFFB7000 - 0xFFFB77FF */
    FLASH_CF_BLOCK_146   = 0x00FB6800,    /*   2KB: 0xFFFB6800 - 0xFFFB6FFF */
    FLASH_CF_BLOCK_147   = 0x00FB6000,    /*   2KB: 0xFFFB6000 - 0xFFFB67FF */
    FLASH_CF_BLOCK_148   = 0x00FB5800,    /*   2KB: 0xFFFB5800 - 0xFFFB5FFF */
    FLASH_CF_BLOCK_149   = 0x00FB5000,    /*   2KB: 0xFFFB5000 - 0xFFFB57FF */
    FLASH_CF_BLOCK_150   = 0x00FB4800,    /*   2KB: 0xFFFB4800 - 0xFFFB4FFF */
    FLASH_CF_BLOCK_151   = 0x00FB4000,    /*   2KB: 0xFFFB4000 - 0xFFFB47FF */
    FLASH_CF_BLOCK_152   = 0x00FB3800,    /*   2KB: 0xFFFB3800 - 0xFFFB3FFF */
    FLASH_CF_BLOCK_153   = 0x00FB3000,    /*   2KB: 0xFFFB3000 - 0xFFFB37FF */
    FLASH_CF_BLOCK_154   = 0x00FB2800,    /*   2KB: 0xFFFB2800 - 0xFFFB2FFF */
    FLASH_CF_BLOCK_155   = 0x00FB2000,    /*   2KB: 0xFFFB2000 - 0xFFFB27FF */
    FLASH_CF_BLOCK_156   = 0x00FB1800,    /*   2KB: 0xFFFB1800 - 0xFFFB1FFF */
    FLASH_CF_BLOCK_157   = 0x00FB1000,    /*   2KB: 0xFFFB1000 - 0xFFFB17FF */
    FLASH_CF_BLOCK_158   = 0x00FB0800,    /*   2KB: 0xFFFB0800 - 0xFFFB0FFF */
    FLASH_CF_BLOCK_159   = 0x00FB0000,    /*   2KB: 0xFFFB0000 - 0xFFFB07FF */
    FLASH_CF_BLOCK_160   = 0x00FAF800,    /*   2KB: 0xFFFAF800 - 0xFFFAFFFF */
    FLASH_CF_BLOCK_161   = 0x00FAF000,    /*   2KB: 0xFFFAF000 - 0xFFFAF7FF */
    FLASH_CF_BLOCK_162   = 0x00FAE800,    /*   2KB: 0xFFFAE800 - 0xFFFAEFFF */
    FLASH_CF_BLOCK_163   = 0x00FAE000,    /*   2KB: 0xFFFAE000 - 0xFFFAE7FF */
    FLASH_CF_BLOCK_164   = 0x00FAD800,    /*   2KB: 0xFFFAD800 - 0xFFFADFFF */
    FLASH_CF_BLOCK_165   = 0x00FAD000,    /*   2KB: 0xFFFAD000 - 0xFFFAD7FF */
    FLASH_CF_BLOCK_166   = 0x00FAC800,    /*   2KB: 0xFFFAC800 - 0xFFFACFFF */
    FLASH_CF_BLOCK_167   = 0x00FAC000,    /*   2KB: 0xFFFAC000 - 0xFFFAC7FF */
    FLASH_CF_BLOCK_168   = 0x00FAB800,    /*   2KB: 0xFFFAB800 - 0xFFFABFFF */
    FLASH_CF_BLOCK_169   = 0x00FAB000,    /*   2KB: 0xFFFAB000 - 0xFFFAB7FF */
    FLASH_CF_BLOCK_170   = 0x00FAA800,    /*   2KB: 0xFFFAA800 - 0xFFFAAFFF */
    FLASH_CF_BLOCK_171   = 0x00FAA000,    /*   2KB: 0xFFFAA000 - 0xFFFAA7FF */
    FLASH_CF_BLOCK_172   = 0x00FA9800,    /*   2KB: 0xFFFA9800 - 0xFFFA9FFF */
    FLASH_CF_BLOCK_173   = 0x00FA9000,    /*   2KB: 0xFFFA9000 - 0xFFFA97FF */
    FLASH_CF_BLOCK_174   = 0x00FA8800,    /*   2KB: 0xFFFA8800 - 0xFFFA8FFF */
    FLASH_CF_BLOCK_175   = 0x00FA8000,    /*   2KB: 0xFFFA8000 - 0xFFFA87FF */
    FLASH_CF_BLOCK_176   = 0x00FA7800,    /*   2KB: 0xFFFA7800 - 0xFFFA7FFF */
    FLASH_CF_BLOCK_177   = 0x00FA7000,    /*   2KB: 0xFFFA7000 - 0xFFFA77FF */
    FLASH_CF_BLOCK_178   = 0x00FA6800,    /*   2KB: 0xFFFA6800 - 0xFFFA6FFF */
    FLASH_CF_BLOCK_179   = 0x00FA6000,    /*   2KB: 0xFFFA6000 - 0xFFFA67FF */
    FLASH_CF_BLOCK_180   = 0x00FA5800,    /*   2KB: 0xFFFA5800 - 0xFFFA5FFF */
    FLASH_CF_BLOCK_181   = 0x00FA5000,    /*   2KB: 0xFFFA5000 - 0xFFFA57FF */
    FLASH_CF_BLOCK_182   = 0x00FA4800,    /*   2KB: 0xFFFA4800 - 0xFFFA4FFF */
    FLASH_CF_BLOCK_183   = 0x00FA4000,    /*   2KB: 0xFFFA4000 - 0xFFFA47FF */
    FLASH_CF_BLOCK_184   = 0x00FA3800,    /*   2KB: 0xFFFA3800 - 0xFFFA3FFF */
    FLASH_CF_BLOCK_185   = 0x00FA3000,    /*   2KB: 0xFFFA3000 - 0xFFFA37FF */
    FLASH_CF_BLOCK_186   = 0x00FA2800,    /*   2KB: 0xFFFA2800 - 0xFFFA2FFF */
    FLASH_CF_BLOCK_187   = 0x00FA2000,    /*   2KB: 0xFFFA2000 - 0xFFFA27FF */
    FLASH_CF_BLOCK_188   = 0x00FA1800,    /*   2KB: 0xFFFA1800 - 0xFFFA1FFF */
    FLASH_CF_BLOCK_189   = 0x00FA1000,    /*   2KB: 0xFFFA1000 - 0xFFFA17FF */
    FLASH_CF_BLOCK_190   = 0x00FA0800,    /*   2KB: 0xFFFA0800 - 0xFFFA0FFF */
    FLASH_CF_BLOCK_191   = 0x00FA0000,    /*   2KB: 0xFFFA0000 - 0xFFFA07FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x7   //   384KB ROM
    FLASH_CF_BLOCK_INVALID = 0x00F9FFFF,
#else
    FLASH_CF_BLOCK_192   = 0x00F9F800,    /*   2KB: 0xFFF9F800 - 0xFFF9FFFF */
    FLASH_CF_BLOCK_193   = 0x00F9F000,    /*   2KB: 0xFFF9F000 - 0xFFF9F7FF */
    FLASH_CF_BLOCK_194   = 0x00F9E800,    /*   2KB: 0xFFF9E800 - 0xFFF9EFFF */
    FLASH_CF_BLOCK_195   = 0x00F9E000,    /*   2KB: 0xFFF9E000 - 0xFFF9E7FF */
    FLASH_CF_BLOCK_196   = 0x00F9D800,    /*   2KB: 0xFFF9D800 - 0xFFF9DFFF */
    FLASH_CF_BLOCK_197   = 0x00F9D000,    /*   2KB: 0xFFF9D000 - 0xFFF9D7FF */
    FLASH_CF_BLOCK_198   = 0x00F9C800,    /*   2KB: 0xFFF9C800 - 0xFFF9CFFF */
    FLASH_CF_BLOCK_199   = 0x00F9C000,    /*   2KB: 0xFFF9C000 - 0xFFF9C7FF */
    FLASH_CF_BLOCK_200   = 0x00F9B800,    /*   2KB: 0xFFF9B800 - 0xFFF9BFFF */
    FLASH_CF_BLOCK_201   = 0x00F9B000,    /*   2KB: 0xFFF9B000 - 0xFFF9B7FF */
    FLASH_CF_BLOCK_202   = 0x00F9A800,    /*   2KB: 0xFFF9A800 - 0xFFF9AFFF */
    FLASH_CF_BLOCK_203   = 0x00F9A000,    /*   2KB: 0xFFF9A000 - 0xFFF9A7FF */
    FLASH_CF_BLOCK_204   = 0x00F99800,    /*   2KB: 0xFFF99800 - 0xFFF99FFF */
    FLASH_CF_BLOCK_205   = 0x00F99000,    /*   2KB: 0xFFF99000 - 0xFFF997FF */
    FLASH_CF_BLOCK_206   = 0x00F98800,    /*   2KB: 0xFFF98800 - 0xFFF98FFF */
    FLASH_CF_BLOCK_207   = 0x00F98000,    /*   2KB: 0xFFF98000 - 0xFFF987FF */
    FLASH_CF_BLOCK_208   = 0x00F97800,    /*   2KB: 0xFFF97800 - 0xFFF97FFF */
    FLASH_CF_BLOCK_209   = 0x00F97000,    /*   2KB: 0xFFF97000 - 0xFFF977FF */
    FLASH_CF_BLOCK_210   = 0x00F96800,    /*   2KB: 0xFFF96800 - 0xFFF96FFF */
    FLASH_CF_BLOCK_211   = 0x00F96000,    /*   2KB: 0xFFF96000 - 0xFFF967FF */
    FLASH_CF_BLOCK_212   = 0x00F95800,    /*   2KB: 0xFFF95800 - 0xFFF95FFF */
    FLASH_CF_BLOCK_213   = 0x00F95000,    /*   2KB: 0xFFF95000 - 0xFFF957FF */
    FLASH_CF_BLOCK_214   = 0x00F94800,    /*   2KB: 0xFFF94800 - 0xFFF94FFF */
    FLASH_CF_BLOCK_215   = 0x00F94000,    /*   2KB: 0xFFF94000 - 0xFFF947FF */
    FLASH_CF_BLOCK_216   = 0x00F93800,    /*   2KB: 0xFFF93800 - 0xFFF93FFF */
    FLASH_CF_BLOCK_217   = 0x00F93000,    /*   2KB: 0xFFF93000 - 0xFFF937FF */
    FLASH_CF_BLOCK_218   = 0x00F92800,    /*   2KB: 0xFFF92800 - 0xFFF92FFF */
    FLASH_CF_BLOCK_219   = 0x00F92000,    /*   2KB: 0xFFF92000 - 0xFFF927FF */
    FLASH_CF_BLOCK_220   = 0x00F91800,    /*   2KB: 0xFFF91800 - 0xFFF91FFF */
    FLASH_CF_BLOCK_221   = 0x00F91000,    /*   2KB: 0xFFF91000 - 0xFFF917FF */
    FLASH_CF_BLOCK_222   = 0x00F90800,    /*   2KB: 0xFFF90800 - 0xFFF90FFF */
    FLASH_CF_BLOCK_223   = 0x00F90000,    /*   2KB: 0xFFF90000 - 0xFFF907FF */
    FLASH_CF_BLOCK_224   = 0x00F8F800,    /*   2KB: 0xFFF8F800 - 0xFFF8FFFF */
    FLASH_CF_BLOCK_225   = 0x00F8F000,    /*   2KB: 0xFFF8F000 - 0xFFF8F7FF */
    FLASH_CF_BLOCK_226   = 0x00F8E800,    /*   2KB: 0xFFF8E800 - 0xFFF8EFFF */
    FLASH_CF_BLOCK_227   = 0x00F8E000,    /*   2KB: 0xFFF8E000 - 0xFFF8E7FF */
    FLASH_CF_BLOCK_228   = 0x00F8D800,    /*   2KB: 0xFFF8D800 - 0xFFF8DFFF */
    FLASH_CF_BLOCK_229   = 0x00F8D000,    /*   2KB: 0xFFF8D000 - 0xFFF8D7FF */
    FLASH_CF_BLOCK_230   = 0x00F8C800,    /*   2KB: 0xFFF8C800 - 0xFFF8CFFF */
    FLASH_CF_BLOCK_231   = 0x00F8C000,    /*   2KB: 0xFFF8C000 - 0xFFF8C7FF */
    FLASH_CF_BLOCK_232   = 0x00F8B800,    /*   2KB: 0xFFF8B800 - 0xFFF8BFFF */
    FLASH_CF_BLOCK_233   = 0x00F8B000,    /*   2KB: 0xFFF8B000 - 0xFFF8B7FF */
    FLASH_CF_BLOCK_234   = 0x00F8A800,    /*   2KB: 0xFFF8A800 - 0xFFF8AFFF */
    FLASH_CF_BLOCK_235   = 0x00F8A000,    /*   2KB: 0xFFF8A000 - 0xFFF8A7FF */
    FLASH_CF_BLOCK_236   = 0x00F89800,    /*   2KB: 0xFFF89800 - 0xFFF89FFF */
    FLASH_CF_BLOCK_237   = 0x00F89000,    /*   2KB: 0xFFF89000 - 0xFFF897FF */
    FLASH_CF_BLOCK_238   = 0x00F88800,    /*   2KB: 0xFFF88800 - 0xFFF88FFF */
    FLASH_CF_BLOCK_239   = 0x00F88000,    /*   2KB: 0xFFF88000 - 0xFFF887FF */
    FLASH_CF_BLOCK_240   = 0x00F87800,    /*   2KB: 0xFFF87800 - 0xFFF87FFF */
    FLASH_CF_BLOCK_241   = 0x00F87000,    /*   2KB: 0xFFF87000 - 0xFFF877FF */
    FLASH_CF_BLOCK_242   = 0x00F86800,    /*   2KB: 0xFFF86800 - 0xFFF86FFF */
    FLASH_CF_BLOCK_243   = 0x00F86000,    /*   2KB: 0xFFF86000 - 0xFFF867FF */
    FLASH_CF_BLOCK_244   = 0x00F85800,    /*   2KB: 0xFFF85800 - 0xFFF85FFF */
    FLASH_CF_BLOCK_245   = 0x00F85000,    /*   2KB: 0xFFF85000 - 0xFFF857FF */
    FLASH_CF_BLOCK_246   = 0x00F84800,    /*   2KB: 0xFFF84800 - 0xFFF84FFF */
    FLASH_CF_BLOCK_247   = 0x00F84000,    /*   2KB: 0xFFF84000 - 0xFFF847FF */
    FLASH_CF_BLOCK_248   = 0x00F83800,    /*   2KB: 0xFFF83800 - 0xFFF83FFF */
    FLASH_CF_BLOCK_249   = 0x00F83000,    /*   2KB: 0xFFF83000 - 0xFFF837FF */
    FLASH_CF_BLOCK_250   = 0x00F82800,    /*   2KB: 0xFFF82800 - 0xFFF82FFF */
    FLASH_CF_BLOCK_251   = 0x00F82000,    /*   2KB: 0xFFF82000 - 0xFFF827FF */
    FLASH_CF_BLOCK_252   = 0x00F81800,    /*   2KB: 0xFFF81800 - 0xFFF81FFF */
    FLASH_CF_BLOCK_253   = 0x00F81000,    /*   2KB: 0xFFF81000 - 0xFFF817FF */
    FLASH_CF_BLOCK_254   = 0x00F80800,    /*   2KB: 0xFFF80800 - 0xFFF80FFF */
    FLASH_CF_BLOCK_255   = 0x00F80000,    /*   2KB: 0xFFF80000 - 0xFFF807FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0x8   //   512KB ROM
    FLASH_CF_BLOCK_INVALID = 0x0000F7FFFF,
#else
    FLASH_CF_BLOCK_256 = 0x00F7F800,    /*   2KB: 0xFFF7F800 - 0xFFF7FFFF */
    FLASH_CF_BLOCK_257 = 0x00F7F000,    /*   2KB: 0xFFF7F000 - 0xFFF7F7FF */
    FLASH_CF_BLOCK_258 = 0x00F7E800,    /*   2KB: 0xFFF7E800 - 0xFFF7EFFF */
    FLASH_CF_BLOCK_259 = 0x00F7E000,    /*   2KB: 0xFFF7E000 - 0xFFF7E7FF */
    FLASH_CF_BLOCK_260 = 0x00F7D800,    /*   2KB: 0xFFF7D800 - 0xFFF7DFFF */
    FLASH_CF_BLOCK_261 = 0x00F7D000,    /*   2KB: 0xFFF7D000 - 0xFFF7D7FF */
    FLASH_CF_BLOCK_262 = 0x00F7C800,    /*   2KB: 0xFFF7C800 - 0xFFF7CFFF */
    FLASH_CF_BLOCK_263 = 0x00F7C000,    /*   2KB: 0xFFF7C000 - 0xFFF7C7FF */
    FLASH_CF_BLOCK_264 = 0x00F7B800,    /*   2KB: 0xFFF7B800 - 0xFFF7BFFF */
    FLASH_CF_BLOCK_265 = 0x00F7B000,    /*   2KB: 0xFFF7B000 - 0xFFF7B7FF */
    FLASH_CF_BLOCK_266 = 0x00F7A800,    /*   2KB: 0xFFF7A800 - 0xFFF7AFFF */
    FLASH_CF_BLOCK_267 = 0x00F7A000,    /*   2KB: 0xFFF7A000 - 0xFFF7A7FF */
    FLASH_CF_BLOCK_268 = 0x00F79800,    /*   2KB: 0xFFF79800 - 0xFFF79FFF */
    FLASH_CF_BLOCK_269 = 0x00F79000,    /*   2KB: 0xFFF79000 - 0xFFF797FF */
    FLASH_CF_BLOCK_270 = 0x00F78800,    /*   2KB: 0xFFF78800 - 0xFFF78FFF */
    FLASH_CF_BLOCK_271 = 0x00F78000,    /*   2KB: 0xFFF78000 - 0xFFF787FF */
    FLASH_CF_BLOCK_272 = 0x00F77800,    /*   2KB: 0xFFF77800 - 0xFFF77FFF */
    FLASH_CF_BLOCK_273 = 0x00F77000,    /*   2KB: 0xFFF77000 - 0xFFF777FF */
    FLASH_CF_BLOCK_274 = 0x00F76800,    /*   2KB: 0xFFF76800 - 0xFFF76FFF */
    FLASH_CF_BLOCK_275 = 0x00F76000,    /*   2KB: 0xFFF76000 - 0xFFF767FF */
    FLASH_CF_BLOCK_276 = 0x00F75800,    /*   2KB: 0xFFF75800 - 0xFFF75FFF */
    FLASH_CF_BLOCK_277 = 0x00F75000,    /*   2KB: 0xFFF75000 - 0xFFF757FF */
    FLASH_CF_BLOCK_278 = 0x00F74800,    /*   2KB: 0xFFF74800 - 0xFFF74FFF */
    FLASH_CF_BLOCK_279 = 0x00F74000,    /*   2KB: 0xFFF74000 - 0xFFF747FF */
    FLASH_CF_BLOCK_280 = 0x00F73800,    /*   2KB: 0xFFF73800 - 0xFFF73FFF */
    FLASH_CF_BLOCK_281 = 0x00F73000,    /*   2KB: 0xFFF73000 - 0xFFF737FF */
    FLASH_CF_BLOCK_282 = 0x00F72800,    /*   2KB: 0xFFF72800 - 0xFFF72FFF */
    FLASH_CF_BLOCK_283 = 0x00F72000,    /*   2KB: 0xFFF72000 - 0xFFF727FF */
    FLASH_CF_BLOCK_284 = 0x00F71800,    /*   2KB: 0xFFF71800 - 0xFFF71FFF */
    FLASH_CF_BLOCK_285 = 0x00F71000,    /*   2KB: 0xFFF71000 - 0xFFF717FF */
    FLASH_CF_BLOCK_286 = 0x00F70800,    /*   2KB: 0xFFF70800 - 0xFFF70FFF */
    FLASH_CF_BLOCK_287 = 0x00F70000,    /*   2KB: 0xFFF70000 - 0xFFF707FF */
    FLASH_CF_BLOCK_288 = 0x00F6F800,    /*   2KB: 0xFFF6F800 - 0xFFF6FFFF */
    FLASH_CF_BLOCK_289 = 0x00F6F000,    /*   2KB: 0xFFF6F000 - 0xFFF6F7FF */
    FLASH_CF_BLOCK_290 = 0x00F6E800,    /*   2KB: 0xFFF6E800 - 0xFFF6EFFF */
    FLASH_CF_BLOCK_291 = 0x00F6E000,    /*   2KB: 0xFFF6E000 - 0xFFF6E7FF */
    FLASH_CF_BLOCK_292 = 0x00F6D800,    /*   2KB: 0xFFF6D800 - 0xFFF6DFFF */
    FLASH_CF_BLOCK_293 = 0x00F6D000,    /*   2KB: 0xFFF6D000 - 0xFFF6D7FF */
    FLASH_CF_BLOCK_294 = 0x00F6C800,    /*   2KB: 0xFFF6C800 - 0xFFF6CFFF */
    FLASH_CF_BLOCK_295 = 0x00F6C000,    /*   2KB: 0xFFF6C000 - 0xFFF6C7FF */
    FLASH_CF_BLOCK_296 = 0x00F6B800,    /*   2KB: 0xFFF6B800 - 0xFFF6BFFF */
    FLASH_CF_BLOCK_297 = 0x00F6B000,    /*   2KB: 0xFFF6B000 - 0xFFF6B7FF */
    FLASH_CF_BLOCK_298 = 0x00F6A800,    /*   2KB: 0xFFF6A800 - 0xFFF6AFFF */
    FLASH_CF_BLOCK_299 = 0x00F6A000,    /*   2KB: 0xFFF6A000 - 0xFFF6A7FF */
    FLASH_CF_BLOCK_300 = 0x00F69800,    /*   2KB: 0xFFF69800 - 0xFFF69FFF */
    FLASH_CF_BLOCK_301 = 0x00F69000,    /*   2KB: 0xFFF69000 - 0xFFF697FF */
    FLASH_CF_BLOCK_302 = 0x00F68800,    /*   2KB: 0xFFF68800 - 0xFFF68FFF */
    FLASH_CF_BLOCK_303 = 0x00F68000,    /*   2KB: 0xFFF68000 - 0xFFF687FF */
    FLASH_CF_BLOCK_304 = 0x00F67800,    /*   2KB: 0xFFF67800 - 0xFFF67FFF */
    FLASH_CF_BLOCK_305 = 0x00F67000,    /*   2KB: 0xFFF67000 - 0xFFF677FF */
    FLASH_CF_BLOCK_306 = 0x00F66800,    /*   2KB: 0xFFF66800 - 0xFFF66FFF */
    FLASH_CF_BLOCK_307 = 0x00F66000,    /*   2KB: 0xFFF66000 - 0xFFF667FF */
    FLASH_CF_BLOCK_308 = 0x00F65800,    /*   2KB: 0xFFF65800 - 0xFFF65FFF */
    FLASH_CF_BLOCK_309 = 0x00F65000,    /*   2KB: 0xFFF65000 - 0xFFF657FF */
    FLASH_CF_BLOCK_310 = 0x00F64800,    /*   2KB: 0xFFF64800 - 0xFFF64FFF */
    FLASH_CF_BLOCK_311 = 0x00F64000,    /*   2KB: 0xFFF64000 - 0xFFF647FF */
    FLASH_CF_BLOCK_312 = 0x00F63800,    /*   2KB: 0xFFF63800 - 0xFFF63FFF */
    FLASH_CF_BLOCK_313 = 0x00F63000,    /*   2KB: 0xFFF63000 - 0xFFF637FF */
    FLASH_CF_BLOCK_314 = 0x00F62800,    /*   2KB: 0xFFF62800 - 0xFFF62FFF */
    FLASH_CF_BLOCK_315 = 0x00F62000,    /*   2KB: 0xFFF62000 - 0xFFF627FF */
    FLASH_CF_BLOCK_316 = 0x00F61800,    /*   2KB: 0xFFF61800 - 0xFFF61FFF */
    FLASH_CF_BLOCK_317 = 0x00F61000,    /*   2KB: 0xFFF61000 - 0xFFF617FF */
    FLASH_CF_BLOCK_318 = 0x00F60800,    /*   2KB: 0xFFF60800 - 0xFFF60FFF */
    FLASH_CF_BLOCK_319 = 0x00F60000,    /*   2KB: 0xFFF60000 - 0xFFF607FF */
    FLASH_CF_BLOCK_320 = 0x00F5F800,    /*   2KB: 0xFFF5F800 - 0xFFF5FFFF */
    FLASH_CF_BLOCK_321 = 0x00F5F000,    /*   2KB: 0xFFF5F000 - 0xFFF5F7FF */
    FLASH_CF_BLOCK_322 = 0x00F5E800,    /*   2KB: 0xFFF5E800 - 0xFFF5EFFF */
    FLASH_CF_BLOCK_323 = 0x00F5E000,    /*   2KB: 0xFFF5E000 - 0xFFF5E7FF */
    FLASH_CF_BLOCK_324 = 0x00F5D800,    /*   2KB: 0xFFF5D800 - 0xFFF5DFFF */
    FLASH_CF_BLOCK_325 = 0x00F5D000,    /*   2KB: 0xFFF5D000 - 0xFFF5D7FF */
    FLASH_CF_BLOCK_326 = 0x00F5C800,    /*   2KB: 0xFFF5C800 - 0xFFF5CFFF */
    FLASH_CF_BLOCK_327 = 0x00F5C000,    /*   2KB: 0xFFF5C000 - 0xFFF5C7FF */
    FLASH_CF_BLOCK_328 = 0x00F5B800,    /*   2KB: 0xFFF5B800 - 0xFFF5BFFF */
    FLASH_CF_BLOCK_329 = 0x00F5B000,    /*   2KB: 0xFFF5B000 - 0xFFF5B7FF */
    FLASH_CF_BLOCK_330 = 0x00F5A800,    /*   2KB: 0xFFF5A800 - 0xFFF5AFFF */
    FLASH_CF_BLOCK_331 = 0x00F5A000,    /*   2KB: 0xFFF5A000 - 0xFFF5A7FF */
    FLASH_CF_BLOCK_332 = 0x00F59800,    /*   2KB: 0xFFF59800 - 0xFFF59FFF */
    FLASH_CF_BLOCK_333 = 0x00F59000,    /*   2KB: 0xFFF59000 - 0xFFF597FF */
    FLASH_CF_BLOCK_334 = 0x00F58800,    /*   2KB: 0xFFF58800 - 0xFFF58FFF */
    FLASH_CF_BLOCK_335 = 0x00F58000,    /*   2KB: 0xFFF58000 - 0xFFF587FF */
    FLASH_CF_BLOCK_336 = 0x00F57800,    /*   2KB: 0xFFF57800 - 0xFFF57FFF */
    FLASH_CF_BLOCK_337 = 0x00F57000,    /*   2KB: 0xFFF57000 - 0xFFF577FF */
    FLASH_CF_BLOCK_338 = 0x00F56800,    /*   2KB: 0xFFF56800 - 0xFFF56FFF */
    FLASH_CF_BLOCK_339 = 0x00F56000,    /*   2KB: 0xFFF56000 - 0xFFF567FF */
    FLASH_CF_BLOCK_340 = 0x00F55800,    /*   2KB: 0xFFF55800 - 0xFFF55FFF */
    FLASH_CF_BLOCK_341 = 0x00F55000,    /*   2KB: 0xFFF55000 - 0xFFF557FF */
    FLASH_CF_BLOCK_342 = 0x00F54800,    /*   2KB: 0xFFF54800 - 0xFFF54FFF */
    FLASH_CF_BLOCK_343 = 0x00F54000,    /*   2KB: 0xFFF54000 - 0xFFF547FF */
    FLASH_CF_BLOCK_344 = 0x00F53800,    /*   2KB: 0xFFF53800 - 0xFFF53FFF */
    FLASH_CF_BLOCK_345 = 0x00F53000,    /*   2KB: 0xFFF53000 - 0xFFF537FF */
    FLASH_CF_BLOCK_346 = 0x00F52800,    /*   2KB: 0xFFF52800 - 0xFFF52FFF */
    FLASH_CF_BLOCK_347 = 0x00F52000,    /*   2KB: 0xFFF52000 - 0xFFF527FF */
    FLASH_CF_BLOCK_348 = 0x00F51800,    /*   2KB: 0xFFF51800 - 0xFFF51FFF */
    FLASH_CF_BLOCK_349 = 0x00F51000,    /*   2KB: 0xFFF51000 - 0xFFF517FF */
    FLASH_CF_BLOCK_350 = 0x00F50800,    /*   2KB: 0xFFF50800 - 0xFFF50FFF */
    FLASH_CF_BLOCK_351 = 0x00F50000,    /*   2KB: 0xFFF50000 - 0xFFF507FF */
    FLASH_CF_BLOCK_352 = 0x00F4F800,    /*   2KB: 0xFFF4F800 - 0xFFF4FFFF */
    FLASH_CF_BLOCK_353 = 0x00F4F000,    /*   2KB: 0xFFF4F000 - 0xFFF4F7FF */
    FLASH_CF_BLOCK_354 = 0x00F4E800,    /*   2KB: 0xFFF4E800 - 0xFFF4EFFF */
    FLASH_CF_BLOCK_355 = 0x00F4E000,    /*   2KB: 0xFFF4E000 - 0xFFF4E7FF */
    FLASH_CF_BLOCK_356 = 0x00F4D800,    /*   2KB: 0xFFF4D800 - 0xFFF4DFFF */
    FLASH_CF_BLOCK_357 = 0x00F4D000,    /*   2KB: 0xFFF4D000 - 0xFFF4D7FF */
    FLASH_CF_BLOCK_358 = 0x00F4C800,    /*   2KB: 0xFFF4C800 - 0xFFF4CFFF */
    FLASH_CF_BLOCK_359 = 0x00F4C000,    /*   2KB: 0xFFF4C000 - 0xFFF4C7FF */
    FLASH_CF_BLOCK_360 = 0x00F4B800,    /*   2KB: 0xFFF4B800 - 0xFFF4BFFF */
    FLASH_CF_BLOCK_361 = 0x00F4B000,    /*   2KB: 0xFFF4B000 - 0xFFF4B7FF */
    FLASH_CF_BLOCK_362 = 0x00F4A800,    /*   2KB: 0xFFF4A800 - 0xFFF4AFFF */
    FLASH_CF_BLOCK_363 = 0x00F4A000,    /*   2KB: 0xFFF4A000 - 0xFFF4A7FF */
    FLASH_CF_BLOCK_364 = 0x00F49800,    /*   2KB: 0xFFF49800 - 0xFFF49FFF */
    FLASH_CF_BLOCK_365 = 0x00F49000,    /*   2KB: 0xFFF49000 - 0xFFF497FF */
    FLASH_CF_BLOCK_366 = 0x00F48800,    /*   2KB: 0xFFF48800 - 0xFFF48FFF */
    FLASH_CF_BLOCK_367 = 0x00F48000,    /*   2KB: 0xFFF48000 - 0xFFF487FF */
    FLASH_CF_BLOCK_368 = 0x00F47800,    /*   2KB: 0xFFF47800 - 0xFFF47FFF */
    FLASH_CF_BLOCK_369 = 0x00F47000,    /*   2KB: 0xFFF47000 - 0xFFF477FF */
    FLASH_CF_BLOCK_370 = 0x00F46800,    /*   2KB: 0xFFF46800 - 0xFFF46FFF */
    FLASH_CF_BLOCK_371 = 0x00F46000,    /*   2KB: 0xFFF46000 - 0xFFF467FF */
    FLASH_CF_BLOCK_372 = 0x00F45800,    /*   2KB: 0xFFF45800 - 0xFFF45FFF */
    FLASH_CF_BLOCK_373 = 0x00F45000,    /*   2KB: 0xFFF45000 - 0xFFF457FF */
    FLASH_CF_BLOCK_374 = 0x00F44800,    /*   2KB: 0xFFF44800 - 0xFFF44FFF */
    FLASH_CF_BLOCK_375 = 0x00F44000,    /*   2KB: 0xFFF44000 - 0xFFF447FF */
    FLASH_CF_BLOCK_376 = 0x00F43800,    /*   2KB: 0xFFF43800 - 0xFFF43FFF */
    FLASH_CF_BLOCK_377 = 0x00F43000,    /*   2KB: 0xFFF43000 - 0xFFF437FF */
    FLASH_CF_BLOCK_378 = 0x00F42800,    /*   2KB: 0xFFF42800 - 0xFFF42FFF */
    FLASH_CF_BLOCK_379 = 0x00F42000,    /*   2KB: 0xFFF42000 - 0xFFF427FF */
    FLASH_CF_BLOCK_380 = 0x00F41800,    /*   2KB: 0xFFF41800 - 0xFFF41FFF */
    FLASH_CF_BLOCK_381 = 0x00F41000,    /*   2KB: 0xFFF41000 - 0xFFF417FF */
    FLASH_CF_BLOCK_382 = 0x00F40800,    /*   2KB: 0xFFF40800 - 0xFFF40FFF */
    FLASH_CF_BLOCK_383 = 0x00F40000,    /*   2KB: 0xFFF40000 - 0xFFF407FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0xa   //   768KB ROM
    FLASH_CF_BLOCK_INVALID = 0x0000F3FFFF,
#else
    FLASH_CF_BLOCK_384 = 0x00F3F800,    /*   2KB: 0xFFF3F800 - 0xFFF3FFFF */
    FLASH_CF_BLOCK_385 = 0x00F3F000,    /*   2KB: 0xFFF3F000 - 0xFFF3F7FF */
    FLASH_CF_BLOCK_386 = 0x00F3E800,    /*   2KB: 0xFFF3E800 - 0xFFF3EFFF */
    FLASH_CF_BLOCK_387 = 0x00F3E000,    /*   2KB: 0xFFF3E000 - 0xFFF3E7FF */
    FLASH_CF_BLOCK_388 = 0x00F3D800,    /*   2KB: 0xFFF3D800 - 0xFFF3DFFF */
    FLASH_CF_BLOCK_389 = 0x00F3D000,    /*   2KB: 0xFFF3D000 - 0xFFF3D7FF */
    FLASH_CF_BLOCK_390 = 0x00F3C800,    /*   2KB: 0xFFF3C800 - 0xFFF3CFFF */
    FLASH_CF_BLOCK_391 = 0x00F3C000,    /*   2KB: 0xFFF3C000 - 0xFFF3C7FF */
    FLASH_CF_BLOCK_392 = 0x00F3B800,    /*   2KB: 0xFFF3B800 - 0xFFF3BFFF */
    FLASH_CF_BLOCK_393 = 0x00F3B000,    /*   2KB: 0xFFF3B000 - 0xFFF3B7FF */
    FLASH_CF_BLOCK_394 = 0x00F3A800,    /*   2KB: 0xFFF3A800 - 0xFFF3AFFF */
    FLASH_CF_BLOCK_395 = 0x00F3A000,    /*   2KB: 0xFFF3A000 - 0xFFF3A7FF */
    FLASH_CF_BLOCK_396 = 0x00F39800,    /*   2KB: 0xFFF39800 - 0xFFF39FFF */
    FLASH_CF_BLOCK_397 = 0x00F39000,    /*   2KB: 0xFFF39000 - 0xFFF397FF */
    FLASH_CF_BLOCK_398 = 0x00F38800,    /*   2KB: 0xFFF38800 - 0xFFF38FFF */
    FLASH_CF_BLOCK_399 = 0x00F38000,    /*   2KB: 0xFFF38000 - 0xFFF387FF */
    FLASH_CF_BLOCK_400 = 0x00F37800,    /*   2KB: 0xFFF37800 - 0xFFF37FFF */
    FLASH_CF_BLOCK_401 = 0x00F37000,    /*   2KB: 0xFFF37000 - 0xFFF377FF */
    FLASH_CF_BLOCK_402 = 0x00F36800,    /*   2KB: 0xFFF36800 - 0xFFF36FFF */
    FLASH_CF_BLOCK_403 = 0x00F36000,    /*   2KB: 0xFFF36000 - 0xFFF367FF */
    FLASH_CF_BLOCK_404 = 0x00F35800,    /*   2KB: 0xFFF35800 - 0xFFF35FFF */
    FLASH_CF_BLOCK_405 = 0x00F35000,    /*   2KB: 0xFFF35000 - 0xFFF357FF */
    FLASH_CF_BLOCK_406 = 0x00F34800,    /*   2KB: 0xFFF34800 - 0xFFF34FFF */
    FLASH_CF_BLOCK_407 = 0x00F34000,    /*   2KB: 0xFFF34000 - 0xFFF347FF */
    FLASH_CF_BLOCK_408 = 0x00F33800,    /*   2KB: 0xFFF33800 - 0xFFF33FFF */
    FLASH_CF_BLOCK_409 = 0x00F33000,    /*   2KB: 0xFFF33000 - 0xFFF337FF */
    FLASH_CF_BLOCK_410 = 0x00F32800,    /*   2KB: 0xFFF32800 - 0xFFF32FFF */
    FLASH_CF_BLOCK_411 = 0x00F32000,    /*   2KB: 0xFFF32000 - 0xFFF327FF */
    FLASH_CF_BLOCK_412 = 0x00F31800,    /*   2KB: 0xFFF31800 - 0xFFF31FFF */
    FLASH_CF_BLOCK_413 = 0x00F31000,    /*   2KB: 0xFFF31000 - 0xFFF317FF */
    FLASH_CF_BLOCK_414 = 0x00F30800,    /*   2KB: 0xFFF30800 - 0xFFF30FFF */
    FLASH_CF_BLOCK_415 = 0x00F30000,    /*   2KB: 0xFFF30000 - 0xFFF307FF */
    FLASH_CF_BLOCK_416 = 0x00F2F800,    /*   2KB: 0xFFF2F800 - 0xFFF2FFFF */
    FLASH_CF_BLOCK_417 = 0x00F2F000,    /*   2KB: 0xFFF2F000 - 0xFFF2F7FF */
    FLASH_CF_BLOCK_418 = 0x00F2E800,    /*   2KB: 0xFFF2E800 - 0xFFF2EFFF */
    FLASH_CF_BLOCK_419 = 0x00F2E000,    /*   2KB: 0xFFF2E000 - 0xFFF2E7FF */
    FLASH_CF_BLOCK_420 = 0x00F2D800,    /*   2KB: 0xFFF2D800 - 0xFFF2DFFF */
    FLASH_CF_BLOCK_421 = 0x00F2D000,    /*   2KB: 0xFFF2D000 - 0xFFF2D7FF */
    FLASH_CF_BLOCK_422 = 0x00F2C800,    /*   2KB: 0xFFF2C800 - 0xFFF2CFFF */
    FLASH_CF_BLOCK_423 = 0x00F2C000,    /*   2KB: 0xFFF2C000 - 0xFFF2C7FF */
    FLASH_CF_BLOCK_424 = 0x00F2B800,    /*   2KB: 0xFFF2B800 - 0xFFF2BFFF */
    FLASH_CF_BLOCK_425 = 0x00F2B000,    /*   2KB: 0xFFF2B000 - 0xFFF2B7FF */
    FLASH_CF_BLOCK_426 = 0x00F2A800,    /*   2KB: 0xFFF2A800 - 0xFFF2AFFF */
    FLASH_CF_BLOCK_427 = 0x00F2A000,    /*   2KB: 0xFFF2A000 - 0xFFF2A7FF */
    FLASH_CF_BLOCK_428 = 0x00F29800,    /*   2KB: 0xFFF29800 - 0xFFF29FFF */
    FLASH_CF_BLOCK_429 = 0x00F29000,    /*   2KB: 0xFFF29000 - 0xFFF297FF */
    FLASH_CF_BLOCK_430 = 0x00F28800,    /*   2KB: 0xFFF28800 - 0xFFF28FFF */
    FLASH_CF_BLOCK_431 = 0x00F28000,    /*   2KB: 0xFFF28000 - 0xFFF287FF */
    FLASH_CF_BLOCK_432 = 0x00F27800,    /*   2KB: 0xFFF27800 - 0xFFF27FFF */
    FLASH_CF_BLOCK_433 = 0x00F27000,    /*   2KB: 0xFFF27000 - 0xFFF277FF */
    FLASH_CF_BLOCK_434 = 0x00F26800,    /*   2KB: 0xFFF26800 - 0xFFF26FFF */
    FLASH_CF_BLOCK_435 = 0x00F26000,    /*   2KB: 0xFFF26000 - 0xFFF267FF */
    FLASH_CF_BLOCK_436 = 0x00F25800,    /*   2KB: 0xFFF25800 - 0xFFF25FFF */
    FLASH_CF_BLOCK_437 = 0x00F25000,    /*   2KB: 0xFFF25000 - 0xFFF257FF */
    FLASH_CF_BLOCK_438 = 0x00F24800,    /*   2KB: 0xFFF24800 - 0xFFF24FFF */
    FLASH_CF_BLOCK_439 = 0x00F24000,    /*   2KB: 0xFFF24000 - 0xFFF247FF */
    FLASH_CF_BLOCK_440 = 0x00F23800,    /*   2KB: 0xFFF23800 - 0xFFF23FFF */
    FLASH_CF_BLOCK_441 = 0x00F23000,    /*   2KB: 0xFFF23000 - 0xFFF237FF */
    FLASH_CF_BLOCK_442 = 0x00F22800,    /*   2KB: 0xFFF22800 - 0xFFF22FFF */
    FLASH_CF_BLOCK_443 = 0x00F22000,    /*   2KB: 0xFFF22000 - 0xFFF227FF */
    FLASH_CF_BLOCK_444 = 0x00F21800,    /*   2KB: 0xFFF21800 - 0xFFF21FFF */
    FLASH_CF_BLOCK_445 = 0x00F21000,    /*   2KB: 0xFFF21000 - 0xFFF217FF */
    FLASH_CF_BLOCK_446 = 0x00F20800,    /*   2KB: 0xFFF20800 - 0xFFF20FFF */
    FLASH_CF_BLOCK_447 = 0x00F20000,    /*   2KB: 0xFFF20000 - 0xFFF207FF */
    FLASH_CF_BLOCK_448 = 0x00F1F800,    /*   2KB: 0xFFF1F800 - 0xFFF1FFFF */
    FLASH_CF_BLOCK_449 = 0x00F1F000,    /*   2KB: 0xFFF1F000 - 0xFFF1F7FF */
    FLASH_CF_BLOCK_450 = 0x00F1E800,    /*   2KB: 0xFFF1E800 - 0xFFF1EFFF */
    FLASH_CF_BLOCK_451 = 0x00F1E000,    /*   2KB: 0xFFF1E000 - 0xFFF1E7FF */
    FLASH_CF_BLOCK_452 = 0x00F1D800,    /*   2KB: 0xFFF1D800 - 0xFFF1DFFF */
    FLASH_CF_BLOCK_453 = 0x00F1D000,    /*   2KB: 0xFFF1D000 - 0xFFF1D7FF */
    FLASH_CF_BLOCK_454 = 0x00F1C800,    /*   2KB: 0xFFF1C800 - 0xFFF1CFFF */
    FLASH_CF_BLOCK_455 = 0x00F1C000,    /*   2KB: 0xFFF1C000 - 0xFFF1C7FF */
    FLASH_CF_BLOCK_456 = 0x00F1B800,    /*   2KB: 0xFFF1B800 - 0xFFF1BFFF */
    FLASH_CF_BLOCK_457 = 0x00F1B000,    /*   2KB: 0xFFF1B000 - 0xFFF1B7FF */
    FLASH_CF_BLOCK_458 = 0x00F1A800,    /*   2KB: 0xFFF1A800 - 0xFFF1AFFF */
    FLASH_CF_BLOCK_459 = 0x00F1A000,    /*   2KB: 0xFFF1A000 - 0xFFF1A7FF */
    FLASH_CF_BLOCK_460 = 0x00F19800,    /*   2KB: 0xFFF19800 - 0xFFF19FFF */
    FLASH_CF_BLOCK_461 = 0x00F19000,    /*   2KB: 0xFFF19000 - 0xFFF197FF */
    FLASH_CF_BLOCK_462 = 0x00F18800,    /*   2KB: 0xFFF18800 - 0xFFF18FFF */
    FLASH_CF_BLOCK_463 = 0x00F18000,    /*   2KB: 0xFFF18000 - 0xFFF187FF */
    FLASH_CF_BLOCK_464 = 0x00F17800,    /*   2KB: 0xFFF17800 - 0xFFF17FFF */
    FLASH_CF_BLOCK_465 = 0x00F17000,    /*   2KB: 0xFFF17000 - 0xFFF177FF */
    FLASH_CF_BLOCK_466 = 0x00F16800,    /*   2KB: 0xFFF16800 - 0xFFF16FFF */
    FLASH_CF_BLOCK_467 = 0x00F16000,    /*   2KB: 0xFFF16000 - 0xFFF167FF */
    FLASH_CF_BLOCK_468 = 0x00F15800,    /*   2KB: 0xFFF15800 - 0xFFF15FFF */
    FLASH_CF_BLOCK_469 = 0x00F15000,    /*   2KB: 0xFFF15000 - 0xFFF157FF */
    FLASH_CF_BLOCK_470 = 0x00F14800,    /*   2KB: 0xFFF14800 - 0xFFF14FFF */
    FLASH_CF_BLOCK_471 = 0x00F14000,    /*   2KB: 0xFFF14000 - 0xFFF147FF */
    FLASH_CF_BLOCK_472 = 0x00F13800,    /*   2KB: 0xFFF13800 - 0xFFF13FFF */
    FLASH_CF_BLOCK_473 = 0x00F13000,    /*   2KB: 0xFFF13000 - 0xFFF137FF */
    FLASH_CF_BLOCK_474 = 0x00F12800,    /*   2KB: 0xFFF12800 - 0xFFF12FFF */
    FLASH_CF_BLOCK_475 = 0x00F12000,    /*   2KB: 0xFFF12000 - 0xFFF127FF */
    FLASH_CF_BLOCK_476 = 0x00F11800,    /*   2KB: 0xFFF11800 - 0xFFF11FFF */
    FLASH_CF_BLOCK_477 = 0x00F11000,    /*   2KB: 0xFFF11000 - 0xFFF117FF */
    FLASH_CF_BLOCK_478 = 0x00F10800,    /*   2KB: 0xFFF10800 - 0xFFF10FFF */
    FLASH_CF_BLOCK_479 = 0x00F10000,    /*   2KB: 0xFFF10000 - 0xFFF107FF */
    FLASH_CF_BLOCK_480 = 0x00F0F800,    /*   2KB: 0xFFF0F800 - 0xFFF0FFFF */
    FLASH_CF_BLOCK_481 = 0x00F0F000,    /*   2KB: 0xFFF0F000 - 0xFFF0F7FF */
    FLASH_CF_BLOCK_482 = 0x00F0E800,    /*   2KB: 0xFFF0E800 - 0xFFF0EFFF */
    FLASH_CF_BLOCK_483 = 0x00F0E000,    /*   2KB: 0xFFF0E000 - 0xFFF0E7FF */
    FLASH_CF_BLOCK_484 = 0x00F0D800,    /*   2KB: 0xFFF0D800 - 0xFFF0DFFF */
    FLASH_CF_BLOCK_485 = 0x00F0D000,    /*   2KB: 0xFFF0D000 - 0xFFF0D7FF */
    FLASH_CF_BLOCK_486 = 0x00F0C800,    /*   2KB: 0xFFF0C800 - 0xFFF0CFFF */
    FLASH_CF_BLOCK_487 = 0x00F0C000,    /*   2KB: 0xFFF0C000 - 0xFFF0C7FF */
    FLASH_CF_BLOCK_488 = 0x00F0B800,    /*   2KB: 0xFFF0B800 - 0xFFF0BFFF */
    FLASH_CF_BLOCK_489 = 0x00F0B000,    /*   2KB: 0xFFF0B000 - 0xFFF0B7FF */
    FLASH_CF_BLOCK_490 = 0x00F0A800,    /*   2KB: 0xFFF0A800 - 0xFFF0AFFF */
    FLASH_CF_BLOCK_491 = 0x00F0A000,    /*   2KB: 0xFFF0A000 - 0xFFF0A7FF */
    FLASH_CF_BLOCK_492 = 0x00F09800,    /*   2KB: 0xFFF09800 - 0xFFF09FFF */
    FLASH_CF_BLOCK_493 = 0x00F09000,    /*   2KB: 0xFFF09000 - 0xFFF097FF */
    FLASH_CF_BLOCK_494 = 0x00F08800,    /*   2KB: 0xFFF08800 - 0xFFF08FFF */
    FLASH_CF_BLOCK_495 = 0x00F08000,    /*   2KB: 0xFFF08000 - 0xFFF087FF */
    FLASH_CF_BLOCK_496 = 0x00F07800,    /*   2KB: 0xFFF07800 - 0xFFF07FFF */
    FLASH_CF_BLOCK_497 = 0x00F07000,    /*   2KB: 0xFFF07000 - 0xFFF077FF */
    FLASH_CF_BLOCK_498 = 0x00F06800,    /*   2KB: 0xFFF06800 - 0xFFF06FFF */
    FLASH_CF_BLOCK_499 = 0x00F06000,    /*   2KB: 0xFFF06000 - 0xFFF067FF */
    FLASH_CF_BLOCK_500 = 0x00F05800,    /*   2KB: 0xFFF05800 - 0xFFF05FFF */
    FLASH_CF_BLOCK_501 = 0x00F05000,    /*   2KB: 0xFFF05000 - 0xFFF057FF */
    FLASH_CF_BLOCK_502 = 0x00F04800,    /*   2KB: 0xFFF04800 - 0xFFF04FFF */
    FLASH_CF_BLOCK_503 = 0x00F04000,    /*   2KB: 0xFFF04000 - 0xFFF047FF */
    FLASH_CF_BLOCK_504 = 0x00F03800,    /*   2KB: 0xFFF03800 - 0xFFF03FFF */
    FLASH_CF_BLOCK_505 = 0x00F03000,    /*   2KB: 0xFFF03000 - 0xFFF037FF */
    FLASH_CF_BLOCK_506 = 0x00F02800,    /*   2KB: 0xFFF02800 - 0xFFF02FFF */
    FLASH_CF_BLOCK_507 = 0x00F02000,    /*   2KB: 0xFFF02000 - 0xFFF027FF */
    FLASH_CF_BLOCK_508 = 0x00F01800,    /*   2KB: 0xFFF01800 - 0xFFF01FFF */
    FLASH_CF_BLOCK_509 = 0x00F01000,    /*   2KB: 0xFFF01000 - 0xFFF017FF */
    FLASH_CF_BLOCK_510 = 0x00F00800,    /*   2KB: 0xFFF00800 - 0xFFF00FFF */
    FLASH_CF_BLOCK_511 = 0x00F00000,    /*   2KB: 0xFFF00000 - 0xFFF007FF */
#if   MCU_CFG_PART_MEMORY_SIZE == 0xb   //   1MB ROM
    FLASH_CF_BLOCK_INVALID = 0x00EFFFFF,
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif


    FLASH_DF_BLOCK_0       = 0x00100000,    /*   128: 0x00100000 - 0x0010007F */
    FLASH_DF_BLOCK_1       = 0x00100080,    /*   128: 0x00100080 - 0x001000FF */
    FLASH_DF_BLOCK_2       = 0x00100100,    /*   128: 0x00100100 - 0x0010017F */
    FLASH_DF_BLOCK_3       = 0x00100180,    /*   128: 0x00100180 - 0x001001FF */
    FLASH_DF_BLOCK_4       = 0x00100200,    /*   128: 0x00100200 - 0x0010027F */
    FLASH_DF_BLOCK_5       = 0x00100280,    /*   128: 0x00100280 - 0x001002FF */
    FLASH_DF_BLOCK_6       = 0x00100300,    /*   128: 0x00100300 - 0x0010037F */
    FLASH_DF_BLOCK_7       = 0x00100380,    /*   128: 0x00100380 - 0x001003FF */
    FLASH_DF_BLOCK_8       = 0x00100400,    /*   128: 0x00100400 - 0x0010047F */
    FLASH_DF_BLOCK_9       = 0x00100480,    /*   128: 0x00100480 - 0x001004FF */
    FLASH_DF_BLOCK_10      = 0x00100500,    /*   128: 0x00100500 - 0x0010057F */
    FLASH_DF_BLOCK_11      = 0x00100580,    /*   128: 0x00100580 - 0x001005FF */
    FLASH_DF_BLOCK_12      = 0x00100600,    /*   128: 0x00100600 - 0x0010067F */
    FLASH_DF_BLOCK_13      = 0x00100680,    /*   128: 0x00100680 - 0x001006FF */
    FLASH_DF_BLOCK_14      = 0x00100700,    /*   128: 0x00100700 - 0x0010077F */
    FLASH_DF_BLOCK_15      = 0x00100780,    /*   128: 0x00100780 - 0x001007FF */
    FLASH_DF_BLOCK_16      = 0x00100800,    /*   128: 0x00100800 - 0x0010087F */
    FLASH_DF_BLOCK_17      = 0x00100880,    /*   128: 0x00100880 - 0x001008FF */
    FLASH_DF_BLOCK_18      = 0x00100900,    /*   128: 0x00100900 - 0x0010097F */
    FLASH_DF_BLOCK_19      = 0x00100980,    /*   128: 0x00100980 - 0x001009FF */
    FLASH_DF_BLOCK_20      = 0x00100A00,    /*   128: 0x00100A00 - 0x00100A7F */
    FLASH_DF_BLOCK_21      = 0x00100A80,    /*   128: 0x00100A80 - 0x00100AFF */
    FLASH_DF_BLOCK_22      = 0x00100B00,    /*   128: 0x00100B00 - 0x00100B7F */
    FLASH_DF_BLOCK_23      = 0x00100B80,    /*   128: 0x00100B80 - 0x00100BFF */
    FLASH_DF_BLOCK_24      = 0x00100C00,    /*   128: 0x00100C00 - 0x00100C7F */
    FLASH_DF_BLOCK_25      = 0x00100C80,    /*   128: 0x00100C80 - 0x00100CFF */
    FLASH_DF_BLOCK_26      = 0x00100D00,    /*   128: 0x00100D00 - 0x00100D7F */
    FLASH_DF_BLOCK_27      = 0x00100D80,    /*   128: 0x00100D80 - 0x00100DFF */
    FLASH_DF_BLOCK_28      = 0x00100E00,    /*   128: 0x00100E00 - 0x00100E7F */
    FLASH_DF_BLOCK_29      = 0x00100E80,    /*   128: 0x00100E80 - 0x00100EFF */
    FLASH_DF_BLOCK_30      = 0x00100F00,    /*   128: 0x00100F00 - 0x00100F7F */
    FLASH_DF_BLOCK_31      = 0x00100F80,    /*   128: 0x00100F80 - 0x00100FFF */
    FLASH_DF_BLOCK_32      = 0x00101000,    /*   128: 0x00101000 - 0x0010107F */
    FLASH_DF_BLOCK_33      = 0x00101080,    /*   128: 0x00101080 - 0x001010FF */
    FLASH_DF_BLOCK_34      = 0x00101100,    /*   128: 0x00101100 - 0x0010117F */
    FLASH_DF_BLOCK_35      = 0x00101180,    /*   128: 0x00101180 - 0x001011FF */
    FLASH_DF_BLOCK_36      = 0x00101200,    /*   128: 0x00101200 - 0x0010127F */
    FLASH_DF_BLOCK_37      = 0x00101280,    /*   128: 0x00101280 - 0x001012FF */
    FLASH_DF_BLOCK_38      = 0x00101300,    /*   128: 0x00101300 - 0x0010137F */
    FLASH_DF_BLOCK_39      = 0x00101380,    /*   128: 0x00101380 - 0x001013FF */
    FLASH_DF_BLOCK_40      = 0x00101400,    /*   128: 0x00101400 - 0x0010147F */
    FLASH_DF_BLOCK_41      = 0x00101480,    /*   128: 0x00101480 - 0x001014FF */
    FLASH_DF_BLOCK_42      = 0x00101500,    /*   128: 0x00101500 - 0x0010157F */
    FLASH_DF_BLOCK_43      = 0x00101580,    /*   128: 0x00101580 - 0x001015FF */
    FLASH_DF_BLOCK_44      = 0x00101600,    /*   128: 0x00101600 - 0x0010167F */
    FLASH_DF_BLOCK_45      = 0x00101680,    /*   128: 0x00101680 - 0x001016FF */
    FLASH_DF_BLOCK_46      = 0x00101700,    /*   128: 0x00101700 - 0x0010177F */
    FLASH_DF_BLOCK_47      = 0x00101780,    /*   128: 0x00101780 - 0x001017FF */
    FLASH_DF_BLOCK_48      = 0x00101800,    /*   128: 0x00101800 - 0x0010187F */
    FLASH_DF_BLOCK_49      = 0x00101880,    /*   128: 0x00101880 - 0x001018FF */
    FLASH_DF_BLOCK_50      = 0x00101900,    /*   128: 0x00101900 - 0x0010197F */
    FLASH_DF_BLOCK_51      = 0x00101980,    /*   128: 0x00101980 - 0x001019FF */
    FLASH_DF_BLOCK_52      = 0x00101A00,    /*   128: 0x00101A00 - 0x00101A7F */
    FLASH_DF_BLOCK_53      = 0x00101A80,    /*   128: 0x00101A80 - 0x00101AFF */
    FLASH_DF_BLOCK_54      = 0x00101B00,    /*   128: 0x00101B00 - 0x00101B7F */
    FLASH_DF_BLOCK_55      = 0x00101B80,    /*   128: 0x00101B80 - 0x00101BFF */
    FLASH_DF_BLOCK_56      = 0x00101C00,    /*   128: 0x00101C00 - 0x00101C7F */
    FLASH_DF_BLOCK_57      = 0x00101C80,    /*   128: 0x00101C80 - 0x00101CFF */
    FLASH_DF_BLOCK_58      = 0x00101D00,    /*   128: 0x00101D00 - 0x00101D7F */
    FLASH_DF_BLOCK_59      = 0x00101D80,    /*   128: 0x00101D80 - 0x00101DFF */
    FLASH_DF_BLOCK_60      = 0x00101E00,    /*   128: 0x00101E00 - 0x00101E7F */
    FLASH_DF_BLOCK_61      = 0x00101E80,    /*   128: 0x00101E80 - 0x00101EFF */
    FLASH_DF_BLOCK_62      = 0x00101F00,    /*   128: 0x00101F00 - 0x00101F7F */
    FLASH_DF_BLOCK_63      = 0x00101F80,    /*   128: 0x00101F80 - 0x00101FFF */
    FLASH_DF_BLOCK_INVALID = 0x00102000     /*   2KB: Can't write beyond 0x00101FFF */

} flash_block_address_t;

typedef struct _rom_block_sizes
{
 uint32_t num_blocks;                     // number of blocks at this size
 uint32_t block_size;                     // Size of each block
}rom_block_sizes_t;

typedef struct _rom_block_info
{
 uint32_t start_addr;                   // starting address for this block section
 uint32_t end_addr;                     // ending (up to and including this) address
 uint16_t block_number;                 // the rom block number for this address queried
 uint32_t thisblock_stAddr;             // the starting address for the above block #
 uint32_t block_size;                   // Size of this block

}rom_block_info_t;

#define NUM_BLOCK_TABLE_ENTRIES 2
static rom_block_sizes_t g_flash_RomBlockSizes[NUM_BLOCK_TABLE_ENTRIES] = {
         512,  2048,    /* 512 blocks of 2K   */
        0,  0
};

/*  According to HW Manual the Max Programming Time for 128 bytes (ROM)
    is 6ms.  This is with a FCLK of 32MHz. The calculation below
    calculates the number of ICLK ticks needed for the timeout delay.
    The 6ms number is adjusted linearly depending on the FCLK frequency.
*/
#define WAIT_MAX_ROM_WRITE \
        ((int32_t)(6000 * (32.0/(MCU_CFG_FCLK_HZ/1000000)))*(MCU_CFG_ICLK_HZ/1000000))

/*  According to HW Manual the Max Programming Time for 8 bytes
    (Data Flash) is 3.2ms.  This is with a FCLK of 32MHz. The calculation
    below calculates the number of ICLK ticks needed for the timeout delay.
    The 3.2ms number is adjusted linearly depending on the FCLK frequency.
*/
#define WAIT_MAX_DF_WRITE \
        ((int32_t)(3200 * (32.0/(MCU_CFG_FCLK_HZ/1000000)))*(MCU_CFG_ICLK_HZ/1000000))

/*  According to HW Manual the Max Blank Check time for 2k bytes
    (Data Flash) is 2.5ms.  This is with a FCLK of 32MHz. The calculation
    below calculates the number of ICLK ticks needed for the timeout delay.
    The 2.5ms number is adjusted linearly depending on the FCLK frequency.
*/
#define WAIT_MAX_BLANK_CHECK \
        ((int32_t)(2500 * (32.0/(MCU_CFG_FCLK_HZ/1000000)))*(MCU_CFG_ICLK_HZ/1000000))
        
/*  According to HW Manual the max timeout value when using the peripheral
    clock notification command is 60us. This is with a FCLK of 32MHz. The 
    calculation below calculates the number of ICLK ticks needed for the 
    timeout delay. The 60us number is adjusted linearly depending on 
    the FCLK frequency.
*/
#define WAIT_MAX_NOTIFY_FCU_CLOCK \
        ((int32_t)(60 * (32.0/(MCU_CFG_FCLK_HZ/1000000)))*(MCU_CFG_ICLK_HZ/1000000))

/*  According to HW Manual the Max Erasure Time for a 2kB block is
    around 60ms.  This is with a FCLK of 32MHz. The calculation below
    calculates the number of ICLK ticks needed for the timeout delay.
    The 60ms number is adjusted linearly depending on the FCLK frequency.
*/
#define WAIT_MAX_ERASE \
        ((int32_t)(60000 * (32.0/(MCU_CFG_FCLK_HZ/1000000)))*(MCU_CFG_ICLK_HZ/1000000))

/******************************************************************************
Error checking
******************************************************************************/
/* FCLK must be between 4MHz and 32MHz. */
#if (MCU_CFG_FCLK_HZ > 32000000) || (MCU_CFG_FCLK_HZ < 4000000)
    #error "ERROR - Flash API - FCLK on RX210 must be between 4MHz and 32MHz!"
#endif

#endif /* _FLASH_API_RX210_H */
